import numpy as np


def policy_eval(policy, env, discount_factor=0.9, theta=0.0001):
    """
    Evaluate a policy given an environment and a full description of the environment's dynamics.

    Args:
        policy: [S, A] shaped matrix representing the policy.
        env: OpenAI env. env.P represents the transition probabilities of the environment.
            env.P[s][a] is a list of transition tuples (prob, next_state, reward, done).
            env.nS is a number of states in the environment.
            env.nA is a number of actions in the environment.
        theta: We stop evaluation once our value function change is less than theta for all states.
        discount_factor: Gamma discount factor.

    Returns:
        Vector of length env.nS representing the value function.
    """
    # Start with a random (all 0) value function
    V = np.zeros(env.env.nS)
    while True:
        delta = 0  # delta = change in value of state from one iteration to next

        for state in range(env.env.nS):  # for all states
            val = 0  # initiate value as 0

            # for all actions/action probabilities
            for action, act_prob in enumerate(policy[state]):
                # transition probabilities,state,rewards of each action
                for prob, next_state, reward, _ in env.env.P[state][action]:
                    val += act_prob * prob * \
                        (reward + discount_factor *
                         V[next_state])  # eqn to calculate
            delta = max(delta, np.abs(val-V[state]))
            V[state] = val
        # break if the change in value is less than the threshold (theta)
        if delta < theta:
            break
    return np.array(V)


def policy_iteration(env, discount_factor=0.9):
    """
    Policy Improvement Algorithm. Iteratively evaluates and improves a policy
    until an optimal policy is found.

    Args:
        env: The OpenAI envrionment.
        policy_eval_fn: Policy Evaluation function that takes 3 arguments:
            policy, env, discount_factor.
        discount_factor: gamma discount factor.

    Returns:
        A tuple (policy, V).
        policy is the optimal policy, a matrix of shape [S, A] where each state s
        contains a valid probability distribution over actions.
        V is the value function for the optimal policy.

    """
    def one_step_lookahead(state, V):
        """
        Helper function to calculate the value for all action in a given state.

        Args:
            state: The state to consider (int)
            V: The value to use as an estimator, Vector of length env.nS

        Returns:
            A vector of length env.nA containing the expected value of each action.
        """
        A = np.zeros(env.env.nA)
        for a in range(env.env.nA):
            for prob, next_state, reward, _ in env.env.P[state][a]:
                A[a] += prob * (reward + discount_factor * V[next_state])
        return A
    # Start with a random policy
    policy = np.ones([env.env.nS, env.env.nA]) / env.env.nA

    iterations = 1

    while True:
        curr_pol_val = policy_eval(
            policy, env, discount_factor)  # eval current policy

        policy_stable = True
        for state in range(env.env.nS):  # for each states
            # best action (Highest prob) under current policy
            chosen_act = np.argmax(policy[state])
            # use one step lookahead to find action values
            act_values = one_step_lookahead(state, curr_pol_val)
            best_act = np.argmax(act_values)  # find best action
            if chosen_act != best_act:
                policy_stable = False  # Greedily find best action
            policy[state] = np.eye(env.env.nA)[best_act]  # update
        if policy_stable:
            return policy, curr_pol_val, iterations

        iterations += 1

    return policy, np.zeros(env.env.nS), iterations



import numpy as np


def get_score(env, policy, terminal_reward=1, episodes=1000, max_steps=10_000):
    # From https://medium.com/analytics-vidhya/solving-the-frozenlake-environment-from-openai-gym-using-value-iteration-5a078dffe438
    misses = 0
    steps_list = []
    for _ in range(episodes):
        observation = env.reset()
        steps = 0
        while steps < max_steps:
            action = policy[observation]
            observation, reward, done, _ = env.step(action)
            steps += 1
            if done and reward == terminal_reward:
                steps_list.append(steps)
                break
            elif done and reward == 0:
                misses += 1
                break

    return np.mean(steps_list), (1 - (misses/episodes))



import numpy as np
from collections import namedtuple


def q_learning(env, learning_rate=0.1, gamma=0.95, epsilon=1.0, max_epsilon=1.0, min_epsilon=0.5, decay_rate=0.0001,
               total_episodes=10_000, max_steps_per_episode=2000, decay=True):
    # From https://python-data-science.readthedocs.io/en/latest/reinforcement.html
    action_size = env.action_space.n
    state_size = env.observation_space.n

    qtable = np.zeros((state_size, action_size))

    EpisodeStats = namedtuple(
        "Stats", ["episode_lengths", "episode_rewards", "episode_deltas"])

    stats = EpisodeStats(
        episode_lengths=np.zeros(total_episodes),
        episode_deltas=np.zeros(total_episodes),
        episode_rewards=np.zeros(total_episodes))

    for episode in range(total_episodes):
        state = env.reset()
        total_rewards = 0

        for step in range(max_steps_per_episode):
            exp_exp_tradeoff = np.random.uniform(0, 1)
            action = None
            # If this number > greater than epsilon --> exploitation (taking the biggest Q value for this state)
            if exp_exp_tradeoff > epsilon:
                action = np.argmax(qtable[state, :])
            # Else doing a random choice --> exploration
            else:
                action = env.action_space.sample()

            # Take the action (a) and observe the outcome state(s') and reward (r)
            new_state, reward, done, _ = env.step(action)
            # Update Q(s,a):= Q(s,a) + lr [R(s,a) + gamma * max Q(s',a') - Q(s,a)]
            # qtable[new_state,:] : all the actions we can take from new state
            delta = (reward + (gamma *
                               np.max(qtable[new_state, :])) - qtable[state, action])

            qtable[state, action] += learning_rate * delta
            total_rewards += reward
            state = new_state

            if done == True:
                stats.episode_deltas[episode] = delta
                stats.episode_rewards[episode] += total_rewards
                stats.episode_lengths[episode] = step
                break

        if decay:
            epsilon = min_epsilon + \
                (max_epsilon - min_epsilon) * \
                np.exp(-decay_rate*episode)

    V = np.reshape(np.max(qtable, axis=1), env.desc.shape)
    policy = np.reshape(np.argmax(qtable, axis=1), env.desc.shape)

    return qtable, policy, V, stats



from setuptools import setup
import setuptools

setup(name='gymutils',
      version='0.0.1',
      url='https://github.com/dennybritz/reinforcement-learning',
      author='Ryan McDermott (forked from Denny Britz)',
      packages=setuptools.find_packages(),
      install_requires=['numpy']
      )



import numpy as np


def value_iteration(env, theta=0.0001, discount_factor=0.9):
    """
    Value Iteration Algorithm.

    Args:
        env: OpenAI env. env.P represents the transition probabilities of the environment.
            env.P[s][a] is a list of transition tuples (prob, next_state, reward, done).
            env.nS is a number of states in the environment.
            env.nA is a number of actions in the environment.
        theta: We stop evaluation once our value function change is less than theta for all states.
        discount_factor: Gamma discount factor.

    Returns:
        A tuple (policy, V) of the optimal policy and the optimal value function.
    """

    def one_step_lookahead(state, V):
        """
        Helper function to calculate the value for all action in a given state.

        Args:
            state: The state to consider (int)
            V: The value to use as an estimator, Vector of length env.nS

        Returns:
            A vector of length env.nA containing the expected value of each action.
        """
        A = np.zeros(env.env.nA)
        for act in range(env.env.nA):
            for prob, next_state, reward, _ in env.env.P[state][act]:
                A[act] += prob * (reward + discount_factor * V[next_state])
        return A

    V = np.zeros(env.env.nS)

    iterations = 1
    while True:
        delta = 0  # checker for improvements across states
        for state in range(env.env.nS):
            act_values = one_step_lookahead(state, V)  # lookahead one step
            best_act_value = np.max(act_values)  # get best action value
            # find max delta across all states
            delta = max(delta, np.abs(best_act_value - V[state]))
            V[state] = best_act_value  # update value to best action value
        if delta < theta:  # if max improvement less than threshold
            break

        iterations += 1

    policy = np.zeros([env.env.nS, env.env.nA])

    for state in range(env.env.nS):  # for all states, create deterministic policy
        act_val = one_step_lookahead(state, V)
        best_action = np.argmax(act_val)
        policy[state][best_action] = 1

    return policy, V, iterations
